/*
 * shohaku Copyright (C) 2005 tomoya nagatani
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 */
package shohaku.configutils;

import java.util.Iterator;

/**
 * コンフィギュレーションリソースを格納する基礎実装。
 */
public abstract class BaseConfigurationResources implements ConfigurationResources {

    /** 保管するコンフィギュレーションリソース。 */
    protected ConfigurationResources lookup;

    /**
     * デフォルトコンストラクタ。
     */
    public BaseConfigurationResources() {
        super();
    }

    /**
     * コンフィギュレーションリソースを格納して初期化します。
     * 
     * @param resources
     *            コンフィギュレーションリソース
     */
    public BaseConfigurationResources(ConfigurationResources resources) {
        super();
        setResources(resources);
    }

    /**
     * 保管するコンフィギュレーションリソースを返却します。
     * 
     * @return コンフィギュレーションリソース
     */
    public ConfigurationResources getResources() {
        return lookup;
    }

    /**
     * コンフィギュレーションリソースを格納します。
     * 
     * @param resources
     *            コンフィギュレーションリソース
     */
    public void setResources(ConfigurationResources resources) {
        if (resources == null) {
            throw new NullPointerException("ConfigurationResources as null");
        }
        lookup = resources;
    }

    /**
     * 全てのリソース識別子を含む反復子を返却します。
     * 
     * @return 全てのリソース識別子を含む反復子
     * @see shohaku.configutils.ConfigurationResources#keyIterator()
     */
    public Iterator keyIterator() {
        return lookup.keyIterator();
    }

    /**
     * リソースを Object 型で取得します。 <br>
     * 指定された識別子が存在しない場合 null を返却します。
     * 
     * @param key
     *            リソース名
     * @return 指定された値
     * @throws NullPointerException
     *             key が null の場合発生する
     * @see shohaku.configutils.ConfigurationResources#getObject(java.lang.String)
     */
    public Object getObject(String key) {
        return lookup.getObject(key);
    }

    /**
     * 指定された文字列型の識別子が示す値を返却します。 <br>
     * 指定された識別子が存在しない場合 defaultValue を返却します。
     * 
     * @param key
     *            リソース識別子
     * @param defaultValue
     *            デフォルト値
     * @return リソース識別子が示す値
     * @throws NullPointerException
     *             key が null の場合発生する
     * @see shohaku.configutils.ConfigurationResources#getObject(java.lang.String, java.lang.Object)
     */
    public Object getObject(String key, Object defaultValue) {
        return lookup.getObject(key, defaultValue);
    }

    /**
     * 識別子を持つリソースが存在するかを返却します。
     * 
     * @param key
     *            リソース識別子
     * @return 指定された識別子が含まれている場合 true
     * @throws NullPointerException
     *             key が null の場合発生する
     * @see shohaku.configutils.ConfigurationResources#containsKey(java.lang.String)
     */
    public boolean containsKey(String key) {
        return lookup.containsKey(key);
    }

    /**
     * 識別子を持つリソースが存在し、指定クラスとキャスト可能な関係に有るを返却します。<br>
     * 値がNullの場合キャスト可能 false 返却します。
     * 
     * @param key
     *            リソース識別子
     * @param type
     *            キャスト可能な関係に有るか検証するクラス
     * @return 指定された識別子が含まれている場合 true
     * @throws NullPointerException
     *             key または type が null の場合発生する
     * @see shohaku.configutils.ConfigurationResources#containsKey(java.lang.String, java.lang.Class)
     */
    public boolean containsKey(String key, Class type) {
        return lookup.containsKey(key, type);
    }

}
