/*
 * shohaku Copyright (C) 2005 tomoya nagatani
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 */
package shohaku.configutils;

import java.io.InputStream;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.Map;

import shohaku.composer.Composer;
import shohaku.composer.CompositeException;
import shohaku.composer.CompositeRule;
import shohaku.composer.Node;
import shohaku.core.beans.BeansFactory;
import shohaku.core.lang.NoSuchResourceException;
import shohaku.core.lang.ObjectCreationException;
import shohaku.core.lang.ResourceLoader;

/**
 * XMLデータで定義された情報を基にコンポーネントを生成する機能を提供します。
 */
public class ComponentFactory {

    /**
     * デフォルトの構成ルールを返却します。
     * 
     * @return デフォルトの構成ルール
     */
    public CompositeRule getDefaultComposeRule() {
        return ConfigUtils.getDefaultComposeRule(ComponentFactory.class);
    }

    /* リソースデータを格納します。 */
    private final Map lookup = new LinkedHashMap();

    /* 使用するComposer。 */
    private final Composer composer = new Composer();

    /* コンポーネントを読み取り初期化します。 */
    private ComponentFactory(InputStream inStream, ClassLoader classLoader) {
        load(inStream, getDefaultComposeRule(), classLoader);
    }

    /* 入力ストリームからリソースデータを読み込む。 */
    private void load(InputStream inStream, CompositeRule rule, ClassLoader classLoader) throws CompositeException {
        //XMLの解析
        composer.setClassLoader(classLoader);
        composer.parse(rule, inStream);
        Node root = composer.getRoot();
        Map lookupValues = new LinkedHashMap();
        initValues(root, lookupValues);
        lookup.putAll(lookupValues);
    }

    /* リソースの初期化および引数に登録をする */
    private void initValues(Node root, Map values) throws CompositeException {
        for (Iterator iter = root.getNodeContext().elementIterator("components"); iter.hasNext();) {
            Node components = (Node) iter.next();
            String name = components.getAttribute("name");
            if (name == null || name.length() == 0) {
                throw new CompositeException("components name is null.");
            }
            initComponentValues(components, name, values);
        }
    }

    /* コンポーネントの初期化および登録をする */
    private void initComponentValues(Node components, String groupName, Map values) throws CompositeException {
        for (Iterator iter = components.getNodeContext().elementIterator("component"); iter.hasNext();) {
            Node component = (Node) iter.next();
            String name = component.getAttribute("name");
            if (name == null || name.length() == 0) {
                throw new CompositeException("component name is null.");
            }
            values.put(groupName + ':' + name, component.getNodeValue());
        }
    }

    /*
     * new component instance
     */

    /**
     * 指定の識別子を持つコンポーネントを生成して返却します。
     * 
     * @param componentId
     *            コンポーネント識別子
     * @return 指定の識別子を持つコンポーネント
     * @throws ObjectCreationException
     *             オブジェクトの生成に失敗した場合に発生します
     */
    public Object getComponent(String componentId) throws ObjectCreationException {
        if (componentId == null) {
            throw new NullPointerException("componentId is null.");
        }
        BeansFactory o = (BeansFactory) lookup.get(componentId);
        if (o == null) {
            throw new ObjectCreationException("componentId is illegal. componentId:" + componentId);
        }
        return o.getInstance();
    }

    /*
     * new factory instance
     */

    /**
     * 定義ファイルからコンポーネントファクトリのインスタンスを生成して返します。
     * 
     * @param resourcePath
     *            定義ファイルへのパス
     * @return コンポーネントファクトリのインスタンス
     * @throws NoSuchResourceException
     *             指定の定義ファイルが発見出来ない場合
     */
    public static ComponentFactory getFactory(String resourcePath) throws NoSuchResourceException {
        return getFactory(resourcePath, ComponentFactory.class.getClassLoader());
    }

    /**
     * 定義ファイルからコンポーネントファクトリのインスタンスを生成して返します。
     * 
     * @param resourcePath
     *            定義ファイルへのパス
     * @return コンポーネントファクトリのインスタンス
     * @param classLoader
     *            生成に使用するクラスローダ
     * @throws NoSuchResourceException
     *             指定の定義ファイルが発見出来ない場合
     */
    public static ComponentFactory getFactory(String resourcePath, ClassLoader classLoader)
            throws NoSuchResourceException {
        return getFactoryForCache(resourcePath, classLoader);
    }

    /* 定義ファイルのパスを基にコンポーネントファクトリのインスタンスを保存します。 */
    private static final Map factoryCache = new HashMap();

    /* 定義ファイルからコンポーネントファクトリのインスタンスを生成して返します。 */
    private static ComponentFactory getFactoryForCache(String resourcePath, ClassLoader classLoader)
            throws NoSuchResourceException {
        if (resourcePath == null) {
            throw new NullPointerException("resourcePath is null.");
        }
        synchronized (factoryCache) {
            ComponentFactory factory = (ComponentFactory) factoryCache.get(resourcePath);
            if (factory != null) {
                return factory;
            } else {
                InputStream inStream = ResourceLoader.getResourceAsStream(resourcePath, classLoader);
                factory = new ComponentFactory(inStream, classLoader);
                factoryCache.put(resourcePath, factory);
            }
            return factory;
        }
    }
}
