/*
 * shohaku Copyright (C) 2005 tomoya nagatani
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 */
package shohaku.configutils;

import java.io.InputStream;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Locale;
import java.util.Map;
import java.util.MissingResourceException;
import java.util.Set;

import shohaku.composer.CompositeException;
import shohaku.composer.CompositeRule;
import shohaku.composer.helpers.XMLCompositeRuleFactory;

/**
 * XMLで定義されたプロパティを用いたリソースバンドルを実装します。
 * <P>
 * <b>XMLリソースバンドルの特徴 </b>
 * <p>
 * XMLリソースバンドルには、ロケール固有のオブジェクトが含まれます。 プログラムで <code>String</code> などのロケール固有のリソースが必要なときは、
 * ユーザのロケールに合ったXMLリソースバンドルからロードできます。 このように、XMLリソースバンドルから、ロケール固有の情報のすべてでなくてもその大部分を切り離すことで、
 * ユーザのロケールにはほとんど依存しないプログラムコードを書くことができます。
 * </p>
 * これにより、以下の特徴を持つプログラムを書くことが可能になります。
 * <UL type=SQUARE>
 * <LI>異なる言語への地域対応、または翻訳が容易
 * <LI>複数のロケールを同時に処理可能
 * <LI>将来サポートするロケールを追加する際の修正が容易
 * </UL>
 * <p>
 * この特徴は java.util.ResourceBundle と同様のものであり、 XMLリソースバンドルの設計は <code>java.util.ResourceBundle</code> の国際化モデルを継承しています。
 * 大きな相違点は、 <code>java.util.ResourceBundle</code> が <code>java.util.Properties</code> を内部的に使用するのに対し、 XMLリソースバンドルは
 * <code>shohaku.configutils.XMLProperties</code> をデータモデルに使用している点です。
 * </p>
 * この機能により <code>java.util.ResourceBundle</code> と同等の柔軟性と、 より複雑なデータ型と複雑な階層データモデルを利用して情報を外部ファイル化する事が出来ます。 <br>
 * XMLリソースバンドルは内部的に <code>shohaku.configutils.XMLProperties</code> を使用するため、 解析ルールを独自にカスタマイズすることが出来ます。 <br>
 * 解析ルールの指定に関しては後記の「XMLリソースバンドル固有の名前規約」で説明します。
 * <P>
 * XMLリソースバンドルの基底名と各ロケールをファミリとして定義する名前規約に関しては <code>java.util.ResourceBundle</code> を参照してください。
 * <P>
 * <b>XMLリソースバンドル固有の名前規約 </b>
 * <p>
 * XMLリソースバンドルには唯一 <code>java.util.ResourceBundle</code> とは別の名前規約が有ります。 <br>
 * XMLリソースバンドルでは基底名を基にして複数のファイルが読取の対象となるため、 同一の基底名に属すファイルは同一の解析ルールが摘要される必要があります。 <br>
 * よって基底名単位で解析ルールを指定する為の以下の配置規約が定義されています。
 * <UL type=SQUARE>
 * <LI>基底名単位で解析ルールを指定する場合のファイル名は '基底名 + _rule.properties' で指定します。
 * </UL>
 * 上記位置に配置しておけばシステムが自動で読み取ります, 配置されていなければデフォルトが使用されます。
 * </p>
 * <P>
 * <b>XMLリソースバンドル固有の問題 </b>
 * <p>
 * XMLリソースバンドルは複数のデータ型に対応する一方で <code>java.util.ResourceBundle</code> と同様に、
 * 上位階層から値が継承されるモデルを採用しているため、上位階層で定義される型とは別の型を下位の階層が定義する可能性が有ります。 <br>
 * この特徴はデータ型の管理を難しくする可能性があります。 この特徴を踏まえて設計および管理を行うことが推奨されます。 <br>
 * DTDファイルを作成して検証を行う、使用する解析ルールの型制約を厳しくする、ドキュメントレベルで厳密に規定する等、 何らかの補助を設ける事で回避することが出来ると考えます。
 * </p>
 */
public class XMLResourceBundle implements ConfigurationResources {

    /*
     * private static final
     */

    /** キャシュの初期容量。 */
    private static final int INITIAL_CACHE_SIZE = 25;

    /** キャシュの負荷係数。 */
    private static final float CACHE_LOAD_FACTOR = (float) 1.0;

    /** XMLResourceBundle のインスタンスをキャシュします。 */
    private static final Hashtable bundleCache = new Hashtable(INITIAL_CACHE_SIZE, CACHE_LOAD_FACTOR);

    /** 候補となるバンドル名の単位でロックを掛ける為のObjectを保管します。 */
    private static final Hashtable lockObjects = new Hashtable(INITIAL_CACHE_SIZE, CACHE_LOAD_FACTOR);

    /*
     * private
     */

    /** プロパティを格納します。 */
    private Map lookupMap;

    /** プロパティ名を格納します。 */
    private Set lookupKeys;

    /** 親のXMLResourceBundleを格納します。 */
    private XMLResourceBundle parent;

    /** ロケールを格納します。 */
    private Locale locale;

    /*
     * constructor
     */

    /**
     * 入力ストリームからプロパティリストを読み込み初期化します。
     * 
     * @param rule
     *            構成ルール
     * @param stream
     *            入力ストリーム
     */
    private XMLResourceBundle(InputStream stream, CompositeRule rule) {
        super();
        XMLProperties properties = new XMLProperties();
        properties.load(stream, rule);
        lookupMap = properties.toMap();
        lookupKeys = lookupMap.keySet();
    }

    /*
     * mathod
     */

    /*
     * access mathod
     */

    /**
     * 全てのプロパティ名を含む反復子を返却します。
     * 
     * @return プロパティ名を含む反復子
     * @see shohaku.configutils.ConfigurationResources#keyIterator()
     */
    public Iterator keyIterator() {
        Iterator parentIterator = (this.parent != null) ? this.parent.keyIterator() : null;
        return new XMLResourceBundleIterator(lookupKeys, parentIterator);
    }

    /**
     * プロパティ名が示す値をString型で返却します。 指定されたキーが存在しない場合Nullを返却します。
     * 
     * @param key
     *            プロパティ名
     * @throws NullPointerException
     *             key が null の場合発生する
     * @throws ClassCastException
     *             取得された値が java.long.String でない場合発生する
     * @return プロパティ名が示すObject
     */
    public String getString(String key) {
        return (String) getObject(key);
    }

    /**
     * プロパティ名が示す値を返却します。 指定されたキーが存在しない場合Nullを返却します。
     * 
     * @param key
     *            プロパティ名
     * @throws NullPointerException
     *             key が null の場合発生する
     * @return プロパティ名が示すObject
     * @see shohaku.configutils.ConfigurationResources#getObject(java.lang.String)
     */
    public Object getObject(String key) {
        if (key == null) {
            throw new NullPointerException("resource key as null");
        }
        Object obj = lookupMap.get(key);
        if (obj == null) {
            if (parent != null) {
                obj = parent.getObject(key);
            }
        }
        return obj;
    }

    /**
     * 指定された文字列型のキーが示す値を返却します。 指定されたキーが存在しない場合 defaultValue を返却します。
     * 
     * @param key
     *            プロパティキー
     * @param defaultValue
     *            プロパティキー
     * @return プロパティキーが示す値
     * @throws NullPointerException
     *             key が null の場合発生する
     * @see shohaku.configutils.ConfigurationResources#getObject(java.lang.String, java.lang.Object)
     */
    public Object getObject(String key, Object defaultValue) {
        Object o = getObject(key);
        if (o == null) {
            o = defaultValue;
        }
        return o;
    }

    /**
     * 親のXMLリソースバンドルも検索対象として、 指定されたキーがプロパティセットに含まれている場合に true を返す。
     * 
     * @param key
     *            プロパティキー
     * @return 指定されたキーが含まれている場合 true
     * @throws NullPointerException
     *             key が null の場合発生する
     * @see shohaku.configutils.ConfigurationResources#containsKey(java.lang.String)
     */
    public boolean containsKey(String key) {
        if (key == null) {
            throw new NullPointerException("resource key as null");
        }
        if (!lookupMap.containsKey(key)) {
            if (parent != null) {
                return lookupMap.containsKey(key);
            }
            return false;
        }
        return true;
    }

    /**
     * 指定されたキーがプロパティセットに含まれていると同時に、 指定されたクラスとキャスト可能な関係に有る場合に true を返す。 値がNullの場合キャスト可能(True)を返却します。
     * 
     * @param key
     *            プロパティキー
     * @param type
     *            キャスト可能な関係に有るか検証するクラス
     * @return 指定されたキーが含まれている場合 true
     * @throws NullPointerException
     *             key または type が null の場合発生する
     * @see shohaku.configutils.ConfigurationResources#containsKey(java.lang.String, java.lang.Class)
     */
    public boolean containsKey(String key, Class type) {
        if (key == null || type == null) {
            throw new NullPointerException("resource key or type as null");
        }
        Object o = getObject(key);
        if (o != null) {
            return type.isInstance(o);
        } else {
            return containsKey(key);
        }
    }

    /**
     * ロケールを返却します。
     * 
     * @return ロケール
     */
    public Locale getLocale() {
        return locale;
    }

    /*
     * private mathod
     */

    /**
     * バンドル名からロケールを取得して格納します。
     * 
     * @param baseName
     *            XMLリソースバンドルの基底名
     * @param bundleName
     *            基底名にロケール情報を付加したバンドル名
     */
    private void setLocale(String baseName, String bundleName) {
        if (baseName.length() == bundleName.length()) {
            locale = new Locale("", "");
        } else if (baseName.length() < bundleName.length()) {
            int pos = baseName.length();
            String temp = bundleName.substring(pos + 1);
            pos = temp.indexOf('_');
            if (pos == -1) {
                locale = new Locale(temp, "", "");
                return;
            }

            String language = temp.substring(0, pos);
            temp = temp.substring(pos + 1);
            pos = temp.indexOf('_');
            if (pos == -1) {
                locale = new Locale(language, temp, "");
                return;
            }

            String country = temp.substring(0, pos);
            temp = temp.substring(pos + 1);

            locale = new Locale(language, country, temp);
        } else {
            //The base name is longer than the bundle name.
            // Something is very wrong
            //with the calling code.
            throw new IllegalArgumentException();
        }
    }

    /**
     * 親のXMLResourceBundleを格納します。
     * 
     * @param parent
     *            親のXMLResourceBundle
     */
    private void setParent(XMLResourceBundle parent) {
        this.parent = parent;
    }

    /*
     * static mathod
     */

    /**
     * 指定された基底名、デフォルトのロケール、および呼び出し側のクラスローダを使用して、 リソースバンドルを取得します。このメソッドを呼び出すことは、次を呼び出すことと同じです。 getBundle(baseName,
     * Locale.getDefault(), XMLResourceBundle.class.getClassLoader()) getClassLoader() が、XMLResourceBundle
     * のセキュリティ特権により実行されることのみが異なります。 検索とインスタンス生成方法の詳細については、getBundle を参照してください。
     * 
     * @param baseName
     *            基底名
     * @return 指定された基底名とデフォルトのロケールのXMLリソースバンドル
     * @throws MissingResourceException
     *             指定された基底名のリソースバンドルが見つからない場合
     */
    public static XMLResourceBundle getBundle(String baseName) {
        return getBundleImpl(baseName, Locale.getDefault(), getDefaultLoader());
    }

    /**
     * 指定された基底名、クラスローダ、およびデフォルトのロケールを使用して、 リソースバンドルを取得します。このメソッドを呼び出すことは、次を呼び出すことと同じです。 getBundle(baseName,
     * Locale.getDefault(), loader) getClassLoader() が、 XMLResourceBundle のセキュリティ特権により実行されることのみが異なります。
     * 検索とインスタンス生成方法の詳細については、getBundle を参照してください。
     * 
     * @param baseName
     *            基底名
     * @param loader
     *            リソースのロード元のクラスローダ
     * @return 指定された基底名とロケールのXMLリソースバンドル
     * @throws MissingResourceException
     *             指定された基底名のリソースバンドルが見つからない場合
     */
    public static XMLResourceBundle getBundle(String baseName, ClassLoader loader) {
        return getBundleImpl(baseName, Locale.getDefault(), loader);
    }

    /**
     * 指定された基底名、ロケール、および呼び出し側のクラスローダを使用して、 リソースバンドルを取得します。このメソッドを呼び出すことは、次を呼び出すことと同じです。 getBundle(baseName, locale,
     * XMLResourceBundle.class.getClassLoader()) getClassLoader() が、XMLResourceBundle のセキュリティ特権により実行されることのみが異なります。
     * 検索とインスタンス生成方法の詳細については、getBundle を参照してください。
     * 
     * @param baseName
     *            基底名
     * @param locale
     *            ロケール
     * @return 指定された基底名とロケールのXMLリソースバンドル
     * @throws MissingResourceException
     *             指定された基底名のリソースバンドルが見つからない場合
     */
    public static XMLResourceBundle getBundle(String baseName, Locale locale) {
        return getBundleImpl(baseName, locale, getDefaultLoader());
    }

    /**
     * 指定された基底名、ロケール、クラスローダを使用して、XMLリソースバンドルを取得します。
     * <p>
     * 理論的には、 <code>getBundle</code> では次の方法を使用して、XMLリソースバンドルの検出および生成を行います。
     * <p>
     * <code>getBundle</code> は、基底名、指定されたロケール、 およびデフォルトのロケール (<CODE>Locale.getDefault</CODE> から取得したロケール) を使用して、
     * <em>候補のバンドル名</em> のシーケンスを生成します。 指定されたロケールの言語、国、およびバリアントがすべて空の文字列の場合、基底名は候補のバンドル名のみになります。 それ以外の場合には、指定したロケール
     * (language1、country1、variant1) およびデフォルトのロケール (language2、country2、variant2) の属性値から、次のシーケンスが生成されます。
     * <ul>
     * <li>baseName + "_" + language1 + "_" + country1 + "_" + variant1
     * <li>baseName + "_" + language1 + "_" + country1
     * <li>baseName + "_" + language1
     * <li>baseName + "_" + language2 + "_" + country2 + "_" + variant2
     * <li>baseName + "_" + language2 + "_" + country2
     * <li>baseName + "_" + language2
     * <li>baseName
     * </ul>
     * <p>
     * 最終コンポーネントが空の文字列の場合、候補のバンドル名は省略されます。 たとえば、country1 が空の文字列の場合、2 番目の候補のバンドル名は省略されます。
     * <p>
     * <code>getBundle</code> は実際のXMLリソースバンドルを「インスタンス化」できる最初の名前を見つけるために、
     * 候補のバンドル名を繰り返し処理します。各候補バンドル名ごとに、XMLリソースバンドルを次のように作成しようとします。
     * <ul>
     * <li>まず、候補のバンドル名を使用して、クラスをロードしようとします。 指定したクラスローダを使用して、XMLプロパティリソースファイルを見つけようとします。
     * すべての「。」文字を「/」文字に置き換え、「。xml」文字列を追加して、候補のバンドル名からパス名を生成します。 <CODE>ClassLoader.getResource</CODE>
     * を使用して、この名前の「リソース」を見つけようとします (<code>getResource</code> の「リソース」とは、XMLリソースバンドルのコンテンツとは無関係であり、
     * ファイルなどのデータのコンテナにすぎません). 「リソース」を見つけた場合、リソースのコンテンツから新しい <CODE>XMLResourceBundle</CODE> インスタンスを作成します。
     * </ul>
     * <p>
     * XMLリソースバンドルが見つからない場合、 <code>MissingResourceException</code> がスローされます。
     * <p>
     * XMLリソースバンドルが見つかったときは、その親連鎖をインスタンス化します。 <code>getBundle</code> は候補のバンドル名を繰り返し処理します。
     * このバンドル名は、XMLリソースバンドルのバンドル名からバリアント、国、および言語を (前に「_」についたものに関して毎回) 連続して削除することによって、取得します。
     * 上述のように、最終コンポーネントが空の文字列である場合、候補のバンドル名は省略されます。 各候補のバンドル名により、上述のように、XMLリソースバンドルをインスタンス化しようとします。
     * インスタンスに成功した場合は常に、前にインスタント化したXMLリソースバンドルに null 以外の親XMLリソースバンドルがない場合は、 前にインスタンス化したXMLリソースバンドルの <CODE>setParent
     * </CODE> メソッドを新しいXMLリソースバンドルにより呼び出します。
     * <p>
     * <code>getBundle</code> の実装はインスタンス化したXMLリソースバンドルをキャッシュし、 同じXMLリソースバンドルのインスタンスを複数回返します。
     * こうした実装では、XMLリソースバンドルおよびその親連鎖の選択が上述のものと互換性がある限り、 XMLリソースバンドルをインスタンス化するシーケンスを変更する場合もあります。
     * <p>
     * <code>baseName</code> 引数は完全指定のクラス名である必要があります。
     * <p>
     * <strong>例: </strong>
     * XMLプロパティファイル、MyResources.xml、MyResources_fr_CH.xml、MyResources_fr.xml、MyResources_en.xml、が提供されています。
     * すべてのファイルのコンテンツが有効です。 デフォルトのロケールは <code>Locale("en", "UK")</code> です。
     * <p>
     * 表示されたロケール引数値を持つ <code>getBundle</code> の呼び出しは、次のソースからXMLリソースバンドルをインスタンス化します。
     * <ul>
     * <li>Locale("fr", "CH"): result MyResources_fr_CH.xml, parent MyResources_fr.xml, parent MyResources.xml
     * <li>Locale("fr", "FR"): result MyResources_fr.xml, parent MyResources.xml
     * <li>Locale("de", "DE"): result MyResources_en.xml, parent MyResources.xml
     * <li>Locale("en", "US"): result MyResources_en.xml, parent MyResources.xml
     * </ul>
     * <p>
     * 
     * @param baseName
     *            基底名
     * @param locale
     *            ロケール
     * @param loader
     *            リソースのロード元のクラスローダ
     * @return 指定された基底名とロケールのXMLリソースバンドル
     * @throws MissingResourceException
     *             指定された基底名のリソースバンドルが見つからない場合
     */
    public static XMLResourceBundle getBundle(String baseName, Locale locale, ClassLoader loader) {
        return getBundleImpl(baseName, locale, loader);
    }

    /*
     * protected static mathod
     */

    /**
     * XMLリソースバンドルを生成し返却します。
     * 
     * @param stream
     *            入力ストリーム
     * @param rule
     *            構成ルール
     * @return 生成されたXMLリソースバンドル
     */
    private static Object createBundle(InputStream stream, CompositeRule rule) {
        return new XMLResourceBundle(stream, rule);
    }

    /*
     * private static mathod
     */

    /**
     * リソースに使用するデフォルトのクラスローダを返却します。
     * 
     * @return デフォルトのクラスローダ
     */
    private static ClassLoader getDefaultLoader() {
        return XMLResourceBundle.class.getClassLoader();
    }

    /**
     * XMLリソースバンドルを取得する処理を実装します。
     * 
     * @param baseName
     *            基底名
     * @param locale
     *            ロケール
     * @param loader
     *            リソースのロード元のクラスローダ
     * @return 指定された基底名とロケールのXMLリソースバンドル
     * @throws MissingResourceException
     *             指定された基底名のリソースバンドルが見つからない場合
     *  
     */
    private static XMLResourceBundle getBundleImpl(String baseName, Locale locale, ClassLoader loader) {

        if (baseName == null || locale == null || loader == null) {
            throw new NullPointerException();
        }

        Object bundle = findBundle(loader, locale, baseName);
        if (bundle == null) {
            throw new MissingResourceException("Can't find bundle for base name", baseName, locale.toString());
        }
        return (XMLResourceBundle) bundle;
    }

    /**
     * XMLリソースバンドルを検索して返却します。 XMLリソースバンドルを発見できなかった場合Nullを返す。
     * 
     * @param loader
     *            リソースのロード元のクラスローダ
     * @param locale
     *            ロケール
     * @param baseName
     *            基底名
     * @return 検索結果のXMLリソースバンドル
     */
    private static Object findBundle(ClassLoader loader, Locale locale, String baseName) {

        Object target = null;

        String[] bundleNames = null;
        Locale defaultLocale = Locale.getDefault();
        if (!defaultLocale.equals(locale)) {
            bundleNames = calculateBundleNames(baseName, locale);
            target = findBundle(loader, baseName, bundleNames);
        }
        if (target == null) {
            bundleNames = calculateBundleNames(baseName, defaultLocale);
            target = findBundle(loader, baseName, bundleNames);
            if (target == null) {
                target = findBundle(loader, baseName, null, baseName);
            }
        }
        return target;
    }

    /**
     * XMLリソースバンドルを検索して返却します。 XMLリソースバンドルを発見できなかった場合Nullを返す。
     * 
     * @param loader
     *            リソースのロード元のクラスローダ
     * @param baseName
     *            基底名
     * @param bundleNames
     *            候補のバンドル名
     * @return 検索結果のXMLリソースバンドル
     */
    private static Object findBundle(ClassLoader loader, String baseName, String[] bundleNames) {

        XMLResourceBundle parent = null;
        XMLResourceBundle target = null;
        XMLResourceBundle bundle = null;

        for (int i = 0; i < bundleNames.length; i++) {
            String bundleName = bundleNames[i];
            if (bundleName == null) {
                continue;
            }
            bundle = findBundle(loader, baseName, parent, bundleName);
            if (bundle != null) {
                parent = bundle;
                if (i != 0) {
                    target = bundle;
                }
            }
        }
        return target;
    }

    /**
     * XMLリソースバンドルを検索して返却します。 XMLリソースバンドルを発見できなかった場合Nullを返す。
     * 
     * @param loader
     *            リソースのロード元のクラスローダ
     * @param baseName
     *            基底名
     * @param parent
     *            親のXMLリソースバンドル
     * @param bundleName
     *            候補のバンドル名
     * @return 検索結果のXMLリソースバンドル
     */
    private static XMLResourceBundle findBundle(ClassLoader loader, String baseName, XMLResourceBundle parent,
            String bundleName) {

        XMLResourceBundle bundle;

        //候補となるバンドル名単位でロックを行う
        synchronized (getLockObject(bundleName)) {
            bundle = (XMLResourceBundle) findBundleInCache(bundleName);

            if (bundle == null) {
                //try loading the bundle via the class
                // loader
                bundle = (XMLResourceBundle) loadBundle(loader, baseName, bundleName);
                if (bundle != null) {
                    // set the bundle's parent and put it in
                    // the cache
                    if (parent != null) {
                        bundle.setParent(parent);
                    }
                    bundle.setLocale(baseName, bundleName);
                    putBundleInCache(bundleName, bundle);
                }
            }
        }
        return bundle;
    }

    /**
     * XMLリソースファイルを読取りXMLリソースバンドルを生成して返却します。 XMLリソースファイルを発見出来なかった場合Nullを返す。
     * 
     * @param loader
     *            リソースのロード元のクラスローダ
     * @param baseName
     *            基底名
     * @param bundleName
     *            候補のバンドル名
     * @return 生成されたXMLリソースバンドル
     */
    private static Object loadBundle(ClassLoader loader, String baseName, String bundleName) {

        //XMLファイルパスに変換します。
        final String resName = bundleName.replace('.', '/') + ".xml";

        //保護領域で所有する全てのアクセス権を使って検索します。
        InputStream stream = getResourceAsStream(loader, resName);

        //読取に成功した場合XMLリソースバンドルを生成し返却します。
        if (stream != null) {
            // make sure it is buffered
            stream = new java.io.BufferedInputStream(stream);
            try {
                return createBundle(stream, findPropertySetRules(loader, baseName));
            } catch (Exception e) {
                throw new CompositeException("XMLリソースファイルの解析に失敗しました。file:" + bundleName, e);
            } finally {
                try {
                    stream.close();
                } catch (Exception e) {
                    //no op
                }
            }
        }
        //発見できなかった場合Nullを返す。
        return null;
    }

    /**
     * リソースを保護領域で所有する全てのアクセス権を使って検索し結果を入力ストリームで返す。
     * 
     * @param loader
     *            リソースのロード元のクラスローダ
     * @param resName
     *            リソース名
     * @return リソースの入力ストリーム
     */
    private static InputStream getResourceAsStream(final ClassLoader loader, final String resName) {
        InputStream stream = (InputStream) java.security.AccessController
                .doPrivileged(new java.security.PrivilegedAction() {
                    public Object run() {
                        if (loader != null) {
                            return loader.getResourceAsStream(resName);
                        } else {
                            return ClassLoader.getSystemResourceAsStream(resName);
                        }
                    }
                });
        return stream;
    }

    /**
     * 基底名のXMLリソースバンドルが利用する構成ルールを検索して返す。
     * 
     * @param loader
     *            リソースのロード元のクラスローダ
     * @param baseName
     *            基底名
     * @return 構成ルール
     */
    private static CompositeRule findPropertySetRules(final ClassLoader loader, final String baseName) {

        //構成ルールファイルパスに変換します。
        final String ruleName = baseName.replace('.', '/') + "_rule.xml";

        //ルール名単位でロックを行う
        synchronized (getLockObject(ruleName)) {
            CompositeRule rule = (CompositeRule) findBundleInCache(ruleName);
            if (rule == null) {

                //保護領域で所有する全てのアクセス権を使って検索します。
                InputStream stream = getResourceAsStream(loader, ruleName);
                if (stream != null) {
                    XMLCompositeRuleFactory factory = new XMLCompositeRuleFactory();
                    factory.setClassLoader(loader);
                    rule = factory.create(stream);
                } else {
                    rule = getDefaultComposeRule();
                }
                putBundleInCache(ruleName, rule);
            }
            return rule;
        }
    }

    /**
     * デフォルトの構成ルールを返却します。
     * 
     * @return デフォルトの構成ルール
     */
    public static CompositeRule getDefaultComposeRule() {
        return ConfigUtils.getDefaultComposeRule(XMLResourceBundle.class);
    }

    /** 検索候補となるバンドル名の最大数。 */
    private static final int MAX_BUNDLES_SEARCHED = 4;

    /** 基底名のインデックス。 */
    private static final int BASE_INDEX = 0;

    /** 言語を持つバンドル名のインデックス。 */
    private static final int LANGUAGE_INDEX = 1;

    /** 国を持つバンドル名のインデックス。 */
    private static final int COUNTRY_INDEX = 2;

    /** バリアントを持つバンドル名のインデックス。 */
    private static final int VARIANT_INDEX = 3;

    /**
     * 検索候補となるバンドル名を返却します。
     * 
     * @param baseName
     *            基底名
     * @param locale
     *            ロケール
     * @return 候補となるバンドル名
     */
    private static String[] calculateBundleNames(String baseName, Locale locale) {

        final String[] result = new String[MAX_BUNDLES_SEARCHED];
        final String language = locale.getLanguage();
        final int languageLength = language.length();
        final String country = locale.getCountry();
        final int countryLength = country.length();
        final String variant = locale.getVariant();
        final int variantLength = variant.length();

        result[BASE_INDEX] = baseName;

        if (languageLength + countryLength + variantLength == 0) {
            //The locale is "", "", "".
            return result;
        }
        final StringBuffer temp = new StringBuffer(baseName);
        temp.append('_');
        temp.append(language);
        if (languageLength > 0) {
            result[LANGUAGE_INDEX] = temp.toString();
        }

        if (countryLength + variantLength == 0) {
            return result;
        }
        temp.append('_');
        temp.append(country);
        if (countryLength > 0) {
            result[COUNTRY_INDEX] = temp.toString();
        }

        if (variantLength == 0) {
            return result;
        }
        temp.append('_');
        temp.append(variant);
        result[VARIANT_INDEX] = temp.toString();

        return result;
    }

    /**
     * リソース単位でロックを掛ける為のObjectを返却します。
     * 
     * @param bundleName
     *            バンドル名
     * @return リソース単位のロックオブジェクト
     */
    private static Object getLockObject(String bundleName) {
        synchronized (lockObjects) {
            Object key = lockObjects.get(bundleName);
            if (key == null) {
                key = new Object();
                lockObjects.put(bundleName, key);
            }
            return key;
        }
    }

    /**
     * キャシュからバンドル名の示すXMLResourceBundleを返却します。
     * 
     * @param bundleName
     *            バンドル名
     * @return XMLResourceBundle
     */
    private static Object findBundleInCache(String bundleName) {
        synchronized (bundleCache) {
            return bundleCache.get(bundleName);
        }
    }

    /**
     * バンドル名の示すXMLResourceBundleをキャシュします。
     * 
     * @param bundleName
     *            バンドル名
     * @param value
     *            XMLResourceBundle
     */
    private static void putBundleInCache(String bundleName, Object value) {
        synchronized (bundleCache) {
            bundleCache.put(bundleName, value);
        }
    }
}
